/**
 * Copyright (C) 2011 Agrosense <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License"); you may
 * not use this file except in compliance with the License. You may obtain a
 * copy of the License at
 *
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package nl.cloudfarming.client.isobus.project;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.io.SyncFailedException;
import java.util.ArrayList;
import java.util.List;
import javax.swing.AbstractAction;
import javax.swing.JOptionPane;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import nl.cloudfarming.client.util.FilePickerUtil;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.Exceptions;
import org.openide.util.NbBundle;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 *
 * @author Gerben Feenstra
 */

@NbBundle.Messages({
    "isobus.filepicker.filefilter.extension=.xml",
    "isobus.filepicker.filefilter.description=Isobus data (*{0})",
    "isobus.filepicker.title=Isobus Import",
    "isobus.filepicker.approve=Import",
    "isobus.error.import.not_an_isobus_file=The selected file could NOT be recognized as an IsoBus taskdata file!",
    "isobus.action.import=Import Isobus"
})
public class ImportIsobusFiles extends AbstractAction implements ActionListener {

    private static final String FILEFILTEREXTENSION = NbBundle.getMessage(ImportIsobusFiles.class, "isobus.filepicker.filefilter.extension");
    private static final String FILEFILTER_DESCRIPTION = NbBundle.getMessage(ImportIsobusFiles.class, "isobus.filepicker.filefilter.description", FILEFILTEREXTENSION);
    private static final String FILEPICKER_TITLE = NbBundle.getMessage(ImportIsobusFiles.class, "isobus.filepicker.title");
    private static final String FILEPICKER_APPROVE = NbBundle.getMessage(ImportIsobusFiles.class, "isobus.filepicker.approve");
    private static final String IMPORT_ERROR = NbBundle.getMessage(ImportIsobusFiles.class, "isobus.error.import.not_an_isobus_file");
    private IsobusProject project;

    public ImportIsobusFiles(IsobusProject project) {
        super(NbBundle.getMessage(ImportIsobusFiles.class, "isobus.action.import"));
        this.project = project;
    }

    /**
     *
     * @param e
     */
    @Override
    public void actionPerformed(ActionEvent e) {

        // Destination rootfolder
        FileObject destinationRoot = project.getTaskfilesFolder(false);
        FileObject destination = null;

        List<FileObject> copyQueue = null;

        // Build the filechooser
        FilePickerUtil.AgroSenseFilePicker agroSenseFilePicker = new FilePickerUtil.AgroSenseFilePicker("Isobus-directory");
        agroSenseFilePicker.setTitle(FILEPICKER_TITLE);
        agroSenseFilePicker.setDefaultWorkingDirectory(new File(IsobusPreference.LAST_DIR.getValue()));
        agroSenseFilePicker.setApproveText(FILEPICKER_APPROVE);
        agroSenseFilePicker.addFileFilter(FILEFILTER_DESCRIPTION, FILEFILTEREXTENSION);

        // Open dialog, retreive selected file
        File selectedFile = agroSenseFilePicker.showOpenDialog();
        if (selectedFile != null) {
            FileObject selectedFileObject = FileUtil.toFileObject(selectedFile);

            // Add files to copy-queue        
            if (selectedFile != null && selectedFile.isFile()) {
                if (isTaskData(selectedFileObject)) {
                    destination = getDestination(destinationRoot, selectedFileObject.getName());
                    copyQueue = getFilesFor(selectedFileObject);
                } else {
                    JOptionPane.showMessageDialog(null, IMPORT_ERROR);
                    //throw new RuntimeException("NOT AN ISOBUS TASKDATA FILE!");
                    // TODO: Let the user know the selected file is NOT a taskdata file more graciously!
                }
            }
            // Copy the shit to project dir
            try {
                if (copyQueue != null && copyQueue.size() > 0) {
                    // TODO: Moved to Util, remove copy method here in 1.0.19 or greater
                    copy(copyQueue, destination);
                }
            } catch (IOException ex) {
                Exceptions.printStackTrace(ex);
            }
        }
    }

    /**
     * Copy queue to sensor project dir.
     *
     * @param copyQueue List of files to copy to the archive dir.
     * @Param destination FileObject to copy the files to.
     * @throws IOException
     * @deprecated Move to Util
     */
    @Deprecated
    public static void copy(List<FileObject> copyQueue, FileObject destination) throws IOException {
        for (FileObject file : copyQueue) {
            try {
                org.openide.filesystems.FileUtil.copyFile(file, destination, file.getName());
            } catch (SyncFailedException ex) {
                //Do nothing file already exists
            }
        }
        copyQueue.clear();
    }
    
    /**
     * Create unique destination. Foldername will be the same as "name" with a
     * number added if it already exists.
     *
     * @param destinationRoot Root of destination
     * @param name Name of the subfolder to create.
     * @return Subfolder destination
     */
    private FileObject getDestination(FileObject destinationRoot, String name) {

        if (destinationRoot.getFileObject(name) == null) {
            try {
                return destinationRoot.createFolder(name);
            } catch (IOException ex) {
                Exceptions.printStackTrace(ex);
            }
        } else {
            int i = 1;
            FileObject destination = null;

            while (destination == null) {
                String numberedName = name + i;
                FileObject folderName = destinationRoot.getFileObject(numberedName);

                // Folder does not exist yet, create it
                if (folderName == null) {
                    try {
                        destination = destinationRoot.createFolder(numberedName);
                    } catch (IOException ex) {
                        Exceptions.printStackTrace(ex);
                    }
                }
                i++;
            }
            return destination;
        }
        return null;
    }

    /**
     * Check if this is really a taskdata file by reading the XML content
     *
     * @param taskData
     * @return
     */
    private boolean isTaskData(FileObject taskData) {
        DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder docBuilder;
        Document doc = null;
        try {
            docBuilder = docBuilderFactory.newDocumentBuilder();
            doc = docBuilder.parse(taskData.getInputStream());
        } catch (ParserConfigurationException | SAXException | IOException ex) {
            Exceptions.printStackTrace(ex);
        }

        NodeList taskDataNode = doc.getElementsByTagName("ISO11783_TaskData");

        if (taskDataNode.getLength() > 0) {
            return true;
        }
        return false;
    }

    /**
     * Return list of files belonging to the specified taskdata fileobject.
     * Timelog filename and Grid filename will be read from the taskdata XML.
     *
     * @param taskDataXml
     * @return List of files belonging to the specified taskdata XML including
     * the taskdata itself.
     */
    private List<FileObject> getFilesFor(FileObject taskDataXml) {
        List<FileObject> files = new ArrayList<>();

        FileObject parent = taskDataXml.getParent();

        DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder docBuilder;
        Document doc = null;
        try {
            docBuilder = docBuilderFactory.newDocumentBuilder();
            doc = docBuilder.parse(taskDataXml.getInputStream());
        } catch (ParserConfigurationException | SAXException | IOException ex) {
            Exceptions.printStackTrace(ex);
        }

        // Timelogs        
        NodeList tglNames = doc.getElementsByTagName("TLG");
        if (tglNames.getLength() > 0) {

            Node tlgNode = null;
            int i = 0;
            while ((tlgNode = tglNames.item(i)) != null) {
                NamedNodeMap attr = tlgNode.getAttributes();
                Node fileNameAttrNode = attr.getNamedItem("A"); // A = Filename
                String timeLogFileName = fileNameAttrNode.getNodeValue();

                FileObject timeLogXmlFile = getFileFrom(parent, timeLogFileName, "xml");
                FileObject timeLogBinFile = getFileFrom(parent, timeLogFileName, "bin");

                if (timeLogXmlFile != null) {
                    files.add(timeLogXmlFile);
                }
                if (timeLogBinFile != null) {
                    files.add(timeLogBinFile);
                }
                i++;
            }

        }

        // Grids
        NodeList gridNames = doc.getElementsByTagName("GRD");
        if (gridNames.getLength() > 0) {
            Node gridNode = null;
            int i = 0;
            while ((gridNode = gridNames.item(i)) != null) {
                NamedNodeMap attr = gridNode.getAttributes();
                Node fileNameAttrNode = attr.getNamedItem("G"); // G = Filename
                String gridFileName = fileNameAttrNode.getNodeValue();
                
                FileObject gridBinFile = getFileFrom(parent, gridFileName, "bin");

                if (gridBinFile != null) {
                    files.add(gridBinFile);
                }
                i++;
            }
        }
        files.add(taskDataXml);
        return files;
    }

    /**
     * Return file specified by filename and extension as a fileobject, ignoring
     * lower/uppercase.
     *
     * @param parentDir Parentdir to search file in
     * @param fileName Name of the file to search for.
     * @return File in parentdir as specified bij filename and extension
     */
    private FileObject getFileFrom(FileObject parentDir, String fileName, String extension) {
        String fileNameUpper = fileName.toUpperCase();
        String fileNameLower = fileName.toLowerCase();
        String extensionUpper = extension.toUpperCase();
        String extensionLower = extension.toLowerCase();

        FileObject returnFile = parentDir.getFileObject(fileNameUpper, extensionUpper);
        if (returnFile == null) {
            //FileUtil.findBrother(parentDir, NAME)
            returnFile = parentDir.getFileObject(fileNameUpper, extensionLower);
        }
        if (returnFile == null) {
            returnFile = parentDir.getFileObject(fileNameLower, extensionUpper);
        }
        if (returnFile == null) {
            returnFile = parentDir.getFileObject(fileNameLower, extensionLower);
        }
        return returnFile;
    }
}
